import sys

from . import Command


class SeedCommand(Command):
    """Command line utility to pre-seed query data for all active runners in a specified date range"""
    
    def process_runner(self, runner):
        """Extend the process_runner method to generate a sample if necessary"""

        super().process_runner(runner)

        if runner['is_scratched'] == False:
            runner.sample.normalized_query_data


def main():
    """Main entry point for seed console script"""

    SeedCommand.main(sys.argv[1:])



import predictive_punter


def test_indexes(provider):
    """The provider should create all necessary database indexes during initialisation"""

    expected_indexes = {
        predictive_punter.Sample:   [
            [('runner_id', 1)]
        ]
    }

    for entity_type in expected_indexes:
        collection = provider.get_database_collection(entity_type)
        index_keys = [index['key'] for index in collection.index_information().values()]
        for expected_index in expected_indexes[entity_type]:
            assert expected_index in index_keys



def calculate_value(race, places):
    """Calculate the value of the race for the winning combination with the specified number of places"""

    value = 0.00

    combinations = race.get_winning_combinations(places)
    for combination in combinations:
        combination_value = 1.00
        for runner in combination:
            combination_value *= runner.starting_price
        value += combination_value - 1.00

    return value


def test_active_runners(race):
    """The active_runners property should return a list containing all non-scratched runners in the race"""

    assert race.active_runners == [runner for runner in race.runners if runner['is_scratched'] is False]


def test_get_winning_combinations(race):
    """The get_winning_combinations method should return a list of tuples of Runners representing all winning combinations for the specified number of places"""

    results = [
        None,
        None,
        None,
        None
    ]
    for runner in race.active_runners:
        if runner.result <= len(results):
            results[runner.result - 1] = runner

    for places in range(1, 5):
        winning_combinations = race.get_winning_combinations(places)
        assert winning_combinations[0] == tuple([results[index] for index in range(places)])


def test_win_value(race):
    """The win_value property should return the sum of the starting prices of each winner less the number of winners"""

    assert race.win_value == calculate_value(race, 1)


def test_exacta_value(race):
    """The exacta_value property should return the sum of the products of the starting prices for the first and second placed runners in each winning combination, less the number of winning combinations"""

    assert race.exacta_value == calculate_value(race, 2)


def test_trifecta_value(race):
    """The trifecta_value property should return the sum of the products of the starting prices for the first, second and third placed runners in each winning combination, less the number of winning combinations"""

    assert race.trifecta_value == calculate_value(race, 3)


def test_first_four_value(race):
    """The first_four_value property should return the sum of the products of the starting prices for the first, second, third and fourth placed runners in each winning combination, less the number of winning combinations"""

    assert race.first_four_value == calculate_value(race, 4)


def test_total_value(race):
    """The total_value property should return the sum of win, exacta, trifecta and first four values"""

    assert race.total_value == race.win_value + race.exacta_value + race.trifecta_value + race.first_four_value



def test_calculate_expected_times(runner):
    """The calculate_expected_times method should return a tuple containing the minimum, maximum and average expected times for the specified performance list"""

    expected_values = tuple([runner.actual_distance / (momentum / runner.actual_weight) for momentum in runner.career.momentums])

    assert runner.calculate_expected_times('career') == expected_values


def test_races_per_year(runner):
    """The races_per_year property should return the total career starts divded by the horse's age as at the race date"""

    assert runner.races_per_year == runner.career.starts / runner.age


def test_sample(runner, sample):
    """The sample property should return the sample for the runner"""

    assert runner.sample['_id'] == sample['_id']



from datetime import datetime, timedelta

import cache_requests
from lxml import html
import predictive_punter
import punters_client
import pymongo
import pytest
import redis
import requests


@pytest.fixture(scope='session')
def database(database_uri):

    database_name = database_uri.split('/')[-1]
    database_client = pymongo.MongoClient(database_uri)
    database_client.drop_database(database_name)
    return database_client.get_default_database()


@pytest.fixture(scope='session')
def database_uri():

    return 'mongodb://localhost:27017/predictive_punter_test'


@pytest.fixture(scope='session')
def provider(database):

    http_client = None
    try:
        http_client = cache_requests.Session(connection=redis.fromurl('redis://localhost:6379/predictive_punter_test'))
    except BaseException:
        try:
            http_client = cache_requests.Session()
        except BaseException:
            http_client = requests.Session()

    html_parser = html.fromstring

    scraper = punters_client.Scraper(http_client, html_parser)

    return predictive_punter.Provider(database, scraper)


@pytest.fixture(scope='session')
def race(provider):

    for meet in provider.get_meets_by_date(datetime(2016, 2, 1)):
        if meet['track'] == 'Kilmore':

            for race in meet.races:
                if race['number'] == 5:

                    return race


@pytest.fixture(scope='session')
def runner(race):

    for runner in race.runners:
        if runner['number'] == 1:

            return runner


@pytest.fixture(scope='session')
def sample(provider, runner):

    return provider.get_sample_by_runner(runner)


@pytest.fixture(scope='session')
def future_race(provider):

    meet = provider.get_meets_by_date(datetime.now() + timedelta(days=1))[0]
    return meet.races[0]



import numpy
import predictive_punter
import sklearn.preprocessing


def test_expected_values(runner):
    """The generate_sample method should return a dictionary containing all expected values"""

    all_results = [runner.result] + [other_runner.result for other_runner in runner.race.active_runners if other_runner['_id'] != runner['_id']]

    raw_query_data = []
    for key in ('barrier', 'number', 'weight'):
        raw_query_data.append(runner[key])
    for key in ('age', 'carrying', 'races_per_year', 'spell', 'up'):
        raw_query_data.append(getattr(runner, key))
    for key1 in ('at_distance', 'at_distance_on_track', 'at_up', 'career', 'last_10', 'last_12_months', 'on_firm', 'on_good', 'on_heavy', 'on_soft', 'on_synthetic', 'on_track', 'on_turf', 'since_rest', 'with_jockey'):
        raw_query_data.extend(runner.calculate_expected_times(key1))
        performance_list = getattr(runner, key1)
        for key2 in ('earnings', 'earnings_potential', 'fourth_pct', 'result_potential', 'roi', 'second_pct', 'starts', 'third_pct', 'win_pct'):
            raw_query_data.append(getattr(performance_list, key2))
        raw_query_data.extend(performance_list.starting_prices)

    expected_values = {
        'raw_query_data':           raw_query_data,
        'imputed_query_data':       None,
        'normalized_query_data':    None,
        'regression_result':        sklearn.preprocessing.normalize(numpy.array(all_results).reshape(1, -1))[0, 0],
        'classification_result':    runner.result if runner.result is not None and 1 <= runner.result <= 4 else 5,
        'weight':                   runner.race.total_value,
        'predictor_version':        predictive_punter.__version__
    }

    generated_values = predictive_punter.Sample.generate_sample(runner)

    for key in expected_values:
        assert generated_values[key] == expected_values[key]



from datetime import timedelta


def dates(date_from, date_to):
    """Iterate over all dates from date_from to date_to (inclusive)"""

    time_delta = timedelta(days=1)
    if date_from > date_to:
        time_delta *= -1

    next_date = date_from
    while (date_from <= date_to and next_date <= date_to) or (date_from > date_to and next_date >= date_to):
        yield next_date
        next_date += time_delta



import racing_data


def calculate_expected_times(self, performance_list_name):
    """Return a tuple containing the minimum, maximum and average expected times based on the specified performance list's momentums"""

    def calculate_expected_time(momentum):
        if momentum is not None:
            return self.actual_distance / (momentum / self.actual_weight)
    
    return tuple([calculate_expected_time(momentum) for momentum in getattr(self, performance_list_name).momentums]) if self.actual_distance is not None else tuple([None, None, None])

racing_data.Runner.calculate_expected_times = calculate_expected_times


@property
def races_per_year(self):
    """Return total number of career starts for the horse divded by the horse's age as at the race date"""

    if self.age is not None and self.age > 0:
        return self.career.starts / self.age

racing_data.Runner.races_per_year = races_per_year


@property
def sample(self):
    """Return the sample for this runner"""

    return self.get_cached_property('sample', self.provider.get_sample_by_runner, self)

racing_data.Runner.sample = sample



import racing_data


@property
def active_runners(self):
    """Return a list of non-scratched runners in the race"""

    def generate_active_runners():
        return [runner for runner in self.runners if runner['is_scratched'] is False]

    return self.get_cached_property('active_runners', generate_active_runners)

racing_data.Race.active_runners = active_runners


def get_winning_combinations(self, places):
    """Return a list of tuples of Runners representing all winning combinations for the specified number of places"""

    if len(self.runners) >= places:

        def get_combinations(results):

            if len(results) > 0:

                combinations = []

                next_combinations = get_combinations(results[1:])
                for item in results[0]:
                    if next_combinations is None:
                        combinations.append([item])
                    else:
                        for next_combination in next_combinations:
                            combinations.append([item] + next_combination)

                dupes = []
                for index in range(len(combinations)):
                    for item in combinations[index]:
                        if len([combo_item for combo_item in combinations[index] if combo_item == item]) > 1:
                            dupes.append(index)
                            break
                for index in sorted(dupes, reverse=True):
                    del combinations[index]

                return combinations

        results = []
        for count in range(places):
            results.append([])

        for runner in self.active_runners:
            if runner.result is not None and runner.result <= len(results):
                results[runner.result - 1].append(runner)

        for index in range(len(results) - 1):
            if len(results[index + 1]) < 1:
                results[index + 1] = list(results[index])

        return [tuple(combination) for combination in get_combinations(results)]

racing_data.Race.get_winning_combinations = get_winning_combinations


def calculate_value(self, places):
    """Return the value of the winning combinations with the specified number of places for the race"""

    value = 0.00

    combinations = self.get_winning_combinations(places)
    if combinations is not None:
        for combination in combinations:
            combination_value = 1.00
            for runner in combination:
                if runner.starting_price is not None:
                    combination_value *= runner.starting_price
            value += combination_value - 1.00

    return value

racing_data.Race.calculate_value = calculate_value


@property
def win_value(self):
    """Return the sum of the starting prices of all winning runners less the number of winning runners"""
    
    return self.calculate_value(1)

racing_data.Race.win_value = win_value


@property
def exacta_value(self):
    """Return the sum of the products of the starting prices of first and second placed runners in all winning combinations, less the number of winning combinations"""
    
    return self.calculate_value(2)

racing_data.Race.exacta_value = exacta_value


@property
def trifecta_value(self):
    """Return the sum of the products of the starting prices of first, second and third placed runners in all winning combinations, less the number of winning combinations"""
    
    return self.calculate_value(3)

racing_data.Race.trifecta_value = trifecta_value


@property
def first_four_value(self):
    """Return the sum of the products of the starting prices of the first, second, third and fourth placed runners in all winning combinations, less the number of winning combinations"""
    
    return self.calculate_value(4)

racing_data.Race.first_four_value = first_four_value


@property
def total_value(self):
    """Return the sum of the win, exacta, trifecta and first four values for this race"""
    
    total_value = 0.0
    for value in (self.win_value, self.exacta_value, self.trifecta_value, self.first_four_value):
        if value is not None:
            total_value += value
    return total_value

racing_data.Race.total_value = total_value



def test_len(sample):
    """The normalized_query_data property should return a list of the same length as raw_query_data"""

    assert len(sample.normalized_query_data) == len(sample['raw_query_data'])


def test_none_values(sample):
    """The normalized_query_data property should return a list that does not contain any None values"""

    assert len([value for value in sample.normalized_query_data if value is None]) == 0


def test_types(sample):
    """The normalizeed_query_data property should return a list of floats"""

    assert isinstance(sample.normalized_query_data, list)
    for item in sample.normalized_query_data:
        assert isinstance(item, float)



import predictive_punter
import pytest


@pytest.fixture(scope='module')
def seed_command(database_uri):

    predictive_punter.SeedCommand.main(['-d', database_uri, '2016-2-1', '2016-2-2'])


def test_samples(database, seed_command):
    """The seed command should populate the database with the expected number of samples"""

    assert database['samples'].count() == database['runners'].count({'is_scratched': False})


def test_values(database, seed_command):
    """The seed command should set normalized query data values for all samples"""

    for sample in database['samples'].find():
        assert sample['normalized_query_data'] is not None



__version__ = '1.0.0a4'


from . import race
from . import runner
from .sample import Sample
from .provider import Provider

from .command import Command
from .scrape import ScrapeCommand
from .seed import SeedCommand



import racing_data

from . import Sample


class Provider(racing_data.Provider):
    """Extend the racing_data Provider class with additional functionality specific to predictive analytics"""

    @property
    def database_indexes(self):
        """Return a dictionary of required database indexes for each entity type"""

        database_indexes = super().database_indexes
        database_indexes[Sample] = [
            [('runner_id', 1)]
        ]

        return database_indexes

    def get_runner_by_sample(self, sample):
        """Get the runner associated with the specified sample"""

        return self.find_one(racing_data.Runner, {'_id': sample['runner_id']}, {'sample': sample})

    def get_sample_by_runner(self, runner):
        """Get the sample for the specified runner"""

        return self.find_or_create_one(Sample, {'runner_id': runner['_id']}, {'runner': runner}, runner['updated_at'], Sample.generate_sample, runner)



from datetime import datetime

from predictive_punter import date_utils


def test_equal():
    """The dates iterator should generate the expected date when date_from == date_to"""

    expected_dates = [
        datetime(2016, 2, 2)
    ]

    actual_dates = [date for date in date_utils.dates(datetime(2016, 2, 2), datetime(2016, 2, 2))]

    assert actual_dates == expected_dates


def test_forward():
    """The dates iterator should generate the expected dates when date_from < date_to"""

    expected_dates = [
        datetime(2016, 2, 1),
        datetime(2016, 2, 2),
        datetime(2016, 2, 3)
    ]

    actual_dates = [date for date in date_utils.dates(datetime(2016, 2, 1), datetime(2016, 2, 3))]

    assert actual_dates == expected_dates


def test_reverse():
    """The dates iterator should generate the expected dates when date_from > date_to"""

    expected_dates = [
        datetime(2016, 2, 3),
        datetime(2016, 2, 2),
        datetime(2016, 2, 1)
    ]

    actual_dates = [date for date in date_utils.dates(datetime(2016, 2, 3), datetime(2016, 2, 1))]

    assert actual_dates == expected_dates



import threading

import numpy
import racing_data
import sklearn.preprocessing

from . import __version__


class Sample(racing_data.Entity):
    """A sample represents a single row of query data for a predictor"""

    normalizer_locks = dict()

    @classmethod
    def generate_sample(cls, runner):
        """Generate a new sample for the specified runner"""

        raw_query_data = []
        for key in ('barrier', 'number', 'weight'):
            raw_query_data.append(runner[key])
        for key in ('age', 'carrying', 'races_per_year', 'spell', 'up'):
            raw_query_data.append(getattr(runner, key))
        for key1 in ('at_distance', 'at_distance_on_track', 'at_up', 'career', 'last_10', 'last_12_months', 'on_firm', 'on_good', 'on_heavy', 'on_soft', 'on_synthetic', 'on_track', 'on_turf', 'since_rest', 'with_jockey'):
            raw_query_data.extend(runner.calculate_expected_times(key1))
            performance_list = getattr(runner, key1)
            for key2 in ('earnings', 'earnings_potential', 'fourth_pct', 'result_potential', 'roi', 'second_pct', 'starts', 'third_pct', 'win_pct'):
                raw_query_data.append(getattr(performance_list, key2))
            raw_query_data.extend(performance_list.starting_prices)

        regression_result = None
        if runner.result is not None:
            all_results = [runner.result] + [other_runner.result for other_runner in runner.race.active_runners if other_runner['_id'] != runner['_id'] and other_runner.result is not None]
            try:
                regression_result = sklearn.preprocessing.normalize(numpy.array(all_results).reshape(1,-1))[0,0]
            except BaseException:
                pass

        return {
            'raw_query_data':           raw_query_data,
            'imputed_query_data':       None,
            'normalized_query_data':    None,
            'regression_result':        regression_result,
            'classification_result':    runner.result if runner.result is not None and 1 <= runner.result <= 4 else 5,
            'weight':                   runner.race.total_value,
            'predictor_version':        __version__
        }

    @classmethod
    def get_normalizer_lock(cls, race):
        """Get the normalizer lock for the specified race"""

        if race['_id'] not in cls.normalizer_locks:
            cls.normalizer_locks[race['_id']] = threading.RLock()

        return cls.normalizer_locks[race['_id']]

    @property
    def imputed_query_data(self):
        """Impute the raw query data alongside the raw query data for all other active runners in the race"""

        if self['imputed_query_data'] is None:

            all_query_data = [list(self['raw_query_data'])] + [runner.sample['raw_query_data'] for runner in self.runner.race.active_runners if runner['_id'] != self.runner['_id']]

            for index in range(len(all_query_data[0])):
                if all_query_data[0][index] is None:
                    other_values = [value[index] for value in all_query_data[1:] if value[index] is not None]
                    if len(other_values) > 0:
                        all_query_data[0][index] = sum(other_values) / len(other_values)
                    else:
                        all_query_data[0][index] = 0.0

            self['imputed_query_data'] = all_query_data[0]
            self.provider.save(self)

        return self['imputed_query_data']

    @property
    def normalized_query_data(self):
        """Normalize the imputed query data alongside the imputed query data for all other active runners in the race"""

        with self.get_normalizer_lock(self.runner.race):

            if self['normalized_query_data'] is None:
                
                all_query_data = numpy.asarray([self.imputed_query_data] + [runner.sample.imputed_query_data for runner in self.runner.race.active_runners if runner['_id'] != self.runner['_id']])
                self['normalized_query_data'] = sklearn.preprocessing.normalize(all_query_data, axis=0).tolist()[0]
                self.provider.save(self)

            return self['normalized_query_data']

    @property
    def has_expired(self):
        """Expire samples sourced from an incompatible predictor version"""

        return self['predictor_version'].split('.')[0] != __version__.split('.')[0]

    @property
    def runner(self):
        """Return the runner associated with this sample"""

        return self.get_cached_property('runner', self.provider.get_runner_by_sample, self)



from codecs import open
from os import path
from setuptools import find_packages, setup


here = path.abspath(path.dirname(__file__))


with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='predictive_punter',
    version='1.0.0a4',
    description='Applying predictive analytics to horse racing via Python',
    long_description=long_description,
    keywords='predictive analytics horse racing',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    url='https://github.com/justjasongreen/predictive_punter',
    author='Jason Green',
    author_email='justjasongreen@gmail.com',
    license='MIT',

    packages=find_packages(exclude=['tests']),
    setup_requires=[],
    install_requires=[
        'cache_requests',
        'lxml',
        'numpy',
        'punters_client',
        'pymongo',
        'racing_data',
        'redis',
        'requests',
        'scikit-learn',
        'scipy'
    ],
    extras_require={
        'dev':  [
            'bumpversion',
            'check-manifest'
        ],
        'test': [
            'tox'
        ]
    },
    package_data={
        'predictive_punter':   []
    },
    data_files=[],
    entry_points={
        'console_scripts':  [
            'scrape=predictive_punter.scrape:main',
            'seed=predictive_punter.seed:main'
        ]
    }
    )



import predictive_punter


def test_persistence(sample, runner, provider):
    """Subsequent identical calls to get_sample_by_runner should return the same Sample object"""

    assert provider.get_sample_by_runner(runner)['_id'] == sample['_id']


def test_provider(sample, provider):
    """All Sample objects should contain a reference to the provider instance from which they were sourced"""

    assert sample.provider == provider


def test_timestamps(sample):
    """All Sample objects should contain timezone aware created/updated at timestamps"""

    for key in ('created_at', 'updated_at'):
        assert sample[key].tzinfo is not None


def test_types(sample):
    """The get_sample_by_runner method should return a Sample object"""

    assert isinstance(sample, predictive_punter.Sample)


def test_updates(future_race, provider):
    """Subsequent identical calls to get_sample_by_runner with a future runner should return the same Sample object updated"""

    old_runner = provider.get_runners_by_race(future_race)[0]
    old_sample = provider.get_sample_by_runner(old_runner)

    new_runner = provider.get_runners_by_race(future_race)[0]
    new_sample = provider.get_sample_by_runner(new_runner)

    assert new_sample['_id'] == old_sample['_id']
    assert new_sample['updated_at'] > old_sample['updated_at']
